/* Esame di Sistemi Operativi - 10 luglio 2003 - Traccia A - Esercizio 2 */

class Cliente extends Thread {
  private int id;
  private Barman barman1;
  private Barman barman2;
  private int numBevandeCocktail = 3; // numero di bevande per cocktail

  public Cliente (int id, Barman barman1, Barman barman2) {
    this.id=id;
    this.barman1 = barman1;
    this.barman2 = barman2;
  }

  public void run () {
    while (true) {

      /* attende tra una richiesta e l'altra */
      try {
	    sleep ( (int)(Math.random()*(2500)+500) );
	  } catch (InterruptedException e) { System.out.println (e); }

      /* sceglie il barman */
      Barman barman;
      double r = Math.random()*2.0;
      if (r < 1.0)
        barman = barman1;
      else
        barman = barman2;

      /* sceglie le bevande del cocktail */
      int cocktail[] = new int[numBevandeCocktail];
      for (int i = 0; i < numBevandeCocktail; i++) {
		boolean giScelta;
		int bevanda;
		do {
          /* sceglie a caso una bevanda */
          bevanda = (int)(Math.random()*Bevande.numBevandeDisponibili);
          /* verifica che non sia uguale ad una delle bevande gi scelte */
          giScelta = false;
          for (int j = 0; j < i && !giScelta; j++)
            if (bevanda==cocktail[j])
              giScelta = true;
	    } while (giScelta);
        cocktail[i] = bevanda;
      }

      /* chiede il cocktail */
      barman.chiediCocktail(id,cocktail);

      /* consuma il cocktail */
      System.out.println ("Il cliente "+id+" consuma il cocktail("+
                           cocktail[0]+","+cocktail[1]+","+cocktail[2]+")");
	}
  }
}


class Barman {
  private int id;
  private Bevande bevande;

  public Barman (int id, Bevande bevande) {
	this.id = id;
	this.bevande = bevande;
  }

  public synchronized void chiediCocktail (int idCliente, int cocktail[]) {
	System.out.println ("Il cliente "+idCliente+" chiede il cocktail("+
	                     cocktail[0]+","+cocktail[1]+","+cocktail[2]+") al barman "+id);

    while(bevande.iniziaPreparazione(cocktail) == false) {
	  System.out.println ("Il barman "+id+" non trova una o piu' bevande per preparare il cocktail("+
		                   cocktail[0]+","+cocktail[1]+","+cocktail[2]+")");
      System.out.println ("Il cliente "+idCliente+" viene messo in attesa");
      try {
        wait();
	  } catch (InterruptedException e) {}
    }

    System.out.println ("Il barman "+id+" inizia a preparare il cocktail("+
                         cocktail[0]+","+cocktail[1]+","+cocktail[2]+") per il cliente "+idCliente);

	/* tempo di preparazione del cocktail */
	try {
      Thread.sleep ( (int)(Math.random()*(1000)+1000) );
	} catch (InterruptedException e) { System.out.println (e); }

    bevande.terminaPreparazione(cocktail);
    System.out.println ("Il barman "+id+" ha finito di preparare il cocktail("+
                         cocktail[0]+","+cocktail[1]+","+cocktail[2]+") per il cliente "+idCliente);

    notifyAll();
  }

}


class Bevande {

  public static final int numBevandeDisponibili = 20;

  /* l'array seguente memorizza la disponibilit corrente delle diverse bevande */
  private boolean bevande[] = new boolean[numBevandeDisponibili];

  public Bevande () {
	for (int i = 0; i < numBevandeDisponibili; i++)
	  bevande[i] = true;
  }

  public synchronized boolean iniziaPreparazione (int cocktail[]) {
    for (int i = 0; i < cocktail.length; i++)
      if (bevande[cocktail[i]] == false)
        return false;
    for (int i = 0; i < cocktail.length; i++)
      bevande[cocktail[i]] = false;
    return true;
  }

  public synchronized void terminaPreparazione (int cocktail[]) {
    for (int i = 0; i < cocktail.length; i++)
      bevande[cocktail[i]] = true;
  }

}


public class Esercizio2A {

  private static final int numClienti = 10;

  public static void main (String args[]) {
    Bevande bevande = new Bevande ();
    Barman b1 = new Barman (0, bevande);
    Barman b2 = new Barman (1, bevande);
    Cliente clienti[] = new Cliente[numClienti];
    for (int i = 0; i < numClienti; i++) {
      clienti[i] = new Cliente (i,b1,b2);
      clienti[i].start();
    }
  }
}